"""Init script for the xdgenvpy module."""

from xdgenvpy.xdgenv import XDG
from xdgenvpy.xdgenv import XDGPackage
from xdgenvpy.xdgenv import XDGPedanticPackage

# Limit namespace clobber when "import *" is used.
__all__ = (
    'XDG',
    'XDGPackage',
    'XDGPedanticPackage',
)



#!/usr/bin/env python3
"""This is the main entry point when xdgenvpy is ran as a Python module."""

import sys
from argparse import ArgumentParser
from logging import Formatter
from logging import StreamHandler
from logging import WARNING
from logging import getLogger
from os import environ
from pathlib import Path
from sys import stderr

import pkg_resources

from xdgenvpy import XDG
from xdgenvpy import XDGPackage

LOG = None


def setup_logger(level=None):
    """
    Sets up a basic logger, mainly used for warning and error messages.  These
    messages will be sent to the process' stderr.  Actual XDG output will be
    written to stdout.

    :param int level: The minimum log level.
    """
    if not level:
        level = WARNING
    formatter = Formatter('%(levelname)s %(message)s')
    handler = StreamHandler(stderr)
    handler.setFormatter(formatter)
    logger = getLogger()
    logger.setLevel(level)
    logger.addHandler(handler)
    return logger


def get_version_str():
    """:return: String representing the program name and version number."""
    return '%(prog)s ' + pkg_resources.require('xdgenvpy')[0].version


def parse_args():
    """
    Parses the process' command line arguments and returns a simple plain-old
    data object.

    :rtype: object
    :return: Plain old data object representing the CLI arguments.
    """
    parser = ArgumentParser(prog='xdg-env',
                            description='Print XDG Base Directories.')
    parser.add_argument('-V',
                        '--version',
                        action='version',
                        version=get_version_str())
    parser.add_argument('variables',
                        metavar='XDG_VAR(s)',
                        type=str,
                        nargs='*',
                        help='The XDG Base Directory variable(s) to print.')
    parser.add_argument('-p',
                        '--package',
                        dest='package',
                        help='Optional package name.')
    parser.add_argument('-d',
                        '--defaults-only',
                        action='store_true',
                        default=False,
                        dest='defaults_only',
                        help='Ignore environment variables, and output default'
                             ' values only.')
    parser.add_argument('-a',
                        '--all',
                        action='store_true',
                        default=False,
                        dest='all',
                        help='Prints all XDG Base Directory environment'
                             ' variables in a shell-sourceable format.')
    return parser.parse_args()


def get_xdg(package):
    """
    Gets an :class:`XDG` instance based on whether or not a package name was
    specified.  If a package name is specified then it will be incorporated into
    the directory values returned from the :class:`XDG` object.

    :param str package: Optional package name.

    :rtype: XDG|XDGPackage
    :return: Returns an XDG object.
    """
    if package:
        xdg = XDGPackage(package)
    else:
        xdg = XDG()
    return xdg


def print_vars(xdg, variables):
    """
    Prints each of the specified XDG variables.  Note that each variable
    specified must be a valid XDG variable, also enumerated in the dictionary
    :code:`XDG_TO_METHOD_MAPPING`.

    For each XDG variable, the associated value will be printed onto separate
    lines on stdout.  Note that logs will be written to stderr, so you may need
    shell-foo to redirect stdout or stderr to different destinations.

    :param XDG xdg: The XDG object to calculate base directories from.
    :param list variables: A sequence of XDG variables to print.
    """
    error_code = 0
    for var in variables:
        if not (str(var).startswith('XDG_') and hasattr(xdg, var)):
            LOG.error('Invalid XDG variable: %s', var)
            error_code |= 1
        else:
            value = getattr(xdg, var)
            if isinstance(value, str):
                print(f'{var}={value}')
            elif isinstance(value, tuple):
                print(f'{var}=' + ':'.join(value))
            elif isinstance(value, Path):
                print(f'{var}=' + str(value))
            else:
                LOG.error('Unexpected type for XDG variable: %s = %s',
                          var,
                          value)
                error_code |= 1
    return error_code


def delete_env(name):
    """
    Deletes the named environment variable from :code:`os.environ` if it exists.

    In theory we could have used :meth:`os.unsetenv` but that is not guaranteed
    to be supported on all platforms.  Rather than futz with that method, we can
    simply delete the variable from the :code:`os.environ` dictionary directly.

    :param str name: The name of the environment variable to unset.
    """
    if name in environ:
        del environ[name]


def main():
    """The main entry point for this 'application'."""
    # pylint: disable=global-statement
    global LOG
    LOG = setup_logger()
    args = parse_args()
    if args.defaults_only:
        delete_env('XDG_DATA_HOME')
        delete_env('XDG_CONFIG_HOME')
        delete_env('XDG_CACHE_HOME')
        delete_env('XDG_RUNTIME_DIR')
        delete_env('XDG_DATA_DIRS')
        delete_env('XDG_CONFIG_DIRS')
    if args.all:
        args.variables = tuple(['XDG_DATA_HOME',
                                'XDG_CONFIG_HOME',
                                'XDG_CACHE_HOME',
                                'XDG_RUNTIME_DIR',
                                'XDG_DATA_DIRS',
                                'XDG_CONFIG_DIRS'])
    xdg = get_xdg(args.package)
    error = print_vars(xdg, args.variables)
    if error:
        sys.exit(error)


if __name__ == '__main__':
    main()



# /usr/bin/env python3

from functools import lru_cache
from glob import glob
from os import getenv
from pathlib import Path
from shutil import rmtree
from subprocess import run
from sys import executable

from setuptools import Command
from setuptools import find_packages
from setuptools import setup

PROJECT_NAME = 'xdgenvpy'
PROJECT_VERSION = '2.3.3'


@lru_cache(maxsize=2)
def read_repo_file(filename):
    """
    Reads the specified file and returns the full contents as a single string.

    :param Path|str filename: The file to read.

    :rtype: str
    :return: The full contents of the specified file.
    """
    filename = str(Path(__file__).resolve().parent.joinpath(filename))
    with open(filename) as f:
        return f.read()


def load_requirements(filename):
    """
    Loads the :code:`requirements.tx` dependency file and returns the
    dependencies as a sequence.

    The file can contain comments that follow the pound symbol (eg. '#').
    Comments will be removed, each line is stripped of leading and trailing
    whitespace, and empty lines are deleted.  Other than these basic
    transformations, the requirements are left intact.

    :param Path|str filename: The requirements file to read.

    :rtype: tuple
    :return: A sequence of requirements.
    """
    reqs = read_repo_file(filename).splitlines()
    reqs = [str(x).strip() for x in reqs]
    reqs = [x[:x.find('#')] for x in reqs if '#' in x]
    reqs = [x for x in reqs if len(x)]
    return tuple(reqs)


def get_xdgenvpy_packages():
    """
    Finds all packages within this project and only returns the production ready
    ones.  Meaning, test packages will not be included.

    :rtype tuple
    :return: A sequence of package names that will be built into the file
            distribution.
    """
    packages = find_packages()
    packages = [p for p in packages if not p.endswith('_test')]
    return tuple(packages)


def run_external(cmd):
    """
    Runs the external commands, and returns a boolean indicating if the sub-
    process exited with a status code of '0'.

    :param list|tuple cmd: The commands to run.

    :rtype boolean
    :return: Flag indicating if the sub-process exit code is '0'.
    """
    print('Running external: ' + str(cmd))
    proc = run(cmd)
    return proc.returncode == 0


def run_setuppy(args):
    """
    Passes the specified arguments into :code:`python3 setup.py ...` and returns
    a boolean indicating if the sub-process exited with a status code of '0'.

    :param list|tuple args: The arguments to pass into :code:`python3 setup.py`.

    :rtype boolean
    :return: Flag indicating if the sub-process exit code is '0'.
    """
    cmd = [executable, 'setup.py']
    cmd.extend(args)
    return run_external(tuple(cmd))


def run_twine(args):
    """
    Passes the specified arguments into :code:`twine ...` and returns a boolean
    indicating if the sub-process exited with a status code of '0'.

    :param list|tuple args: The arguments to pass into :code:`twine`.

    :rtype boolean
    :return: Flag indicating if the sub-process exit code is '0'.
    """
    cmd = ['twine']
    cmd.extend(args)
    return run_external(tuple(cmd))


def fail(msg):
    """
    :raises RuntimeError to help make readable code by chaining function calls.
    """
    raise RuntimeError(msg)


class CleanCommand(Command):
    """
    A custom clean command that removes any intermediate build directories.

    :param str description: A short description describing the command.
    :param list user_options: Sequence of CLI options for the command.
    """

    description = 'Custom clean command that forcefully removes build, dist,' \
                  ' and other similar directories.'
    user_options = []

    def __init__(self, *args, **kwargs):
        """Initialized the custom clean command with a list of directories."""
        super(CleanCommand, self).__init__(*args, **kwargs)
        project_path = Path(__file__).resolve().parent
        self._clean_paths = {
            '.coverage',
            '.pytest_cache',
            'build',
            'coverage.xml',
            'dist',
            'man/man1/xdg-env.1.gz',
            'man/man3/xdgenvpy.3.gz',
            PROJECT_NAME + '.egg-info',
        }
        self._clean_paths = {project_path.joinpath(p)
                             for p in self._clean_paths}
        self._clean_paths = {d for d in self._clean_paths if d.exists()}

    def initialize_options(self):
        """Unused, but required when implementing :class:`Command`."""
        pass

    def finalize_options(self):
        """Unused, but required when implementing :class:`Command`."""
        pass

    def run(self):
        """Removes all of the intermediate build directories."""
        for path in self._clean_paths:
            if Path(path).is_file():
                print(f'removing file {path}')
                path.unlink()
            else:
                print(f'removing directory {path}')
                rmtree(path)


class BuildManPagesCommand(Command):
    """
    A custom command that builds the custom manpages.

    To view the man-pages locally, both raw and compressed, used the
    :code:`man -l <localfile>` command.

    :param str description A short description describing the command.
    :param list user_options Sequence of CLI options for the command.
    """

    description = 'Builds the manpages.'
    user_options = []

    def __init__(self, *args, **kwargs):
        """
        Initializes the command, which includes building new source and binary
        distribution files.
        """
        super(BuildManPagesCommand, self).__init__(*args, **kwargs)

    def initialize_options(self):
        """Unused, but required when implementing :class:`Command`."""
        pass

    def finalize_options(self):
        """Unused, but required when implementing :class:`Command`."""
        pass

    @staticmethod
    def run():
        """Builds the man-pages.."""
        files = tuple(glob('./man/man[0-9]/*.[0-9]'))
        # TODO How to make this work for Windows?
        run_external(('gzip',
                      '--verbose',
                      '--keep',
                      '--best',
                      '--force',
                      *files))

    @staticmethod
    def get_data_files():
        """Finds the files required by setuptool's :code:`data_files` prop."""
        return [
            ('man/man1', glob('man/man1/*.gz')),
            ('man/man3', glob('man/man3/*.gz')),
        ]


class TwineCheckCommand(Command):
    """
    A custom command that is used to validate built distribution packages before
    publishing them to public repositories.

    :param str description A short description describing the command.
    :param list user_options Sequence of CLI options for the command.
    """

    description = 'Validates built distributions.'
    user_options = []

    def __init__(self, *args, **kwargs):
        """
        Initializes the command, which includes building new source and binary
        distribution files.
        """
        super(TwineCheckCommand, self).__init__(*args, **kwargs)
        run_setuppy(['clean', 'sdist', 'bdist_wheel'])

    def initialize_options(self):
        """Unused, but required when implementing :class:`Command`."""
        pass

    def finalize_options(self):
        """Unused, but required when implementing :class:`Command`."""
        pass

    @staticmethod
    def run():
        """Performs the validation of the source and binary distributions."""
        run_twine(['check', 'dist/*']) \
            or fail('Could not check packages in dist/*')


class PublishCommand(Command):
    """
    A custom command that is used to easily publish new source and binary
    (wheel) distributions to the public PyPi repository.

    This command takes username and password credentials as arguments.  If they
    are not set, the command defaults to the :code:`$PYPI_USERNAME` and
    :code:`$PYPI_PASSWORD` environment variables.  If neither the the CLI
    arguments nor the environment variables are set, then the command will fail
    to publish the distribution files.

    :param str description A short description describing the command.
    :param list user_options Sequence of CLI options for the command.
    """

    description = 'Publishes built distributions to PyPi'
    user_options = [
        ('username=', 'u', 'The username to the PyPi repository. '
                           'Defaults to the $PYPI_USERNAME env variable.'),
        ('password=', 'p', 'The password to the PyPi repository. '
                           'Defaults to the $PYPI_PASSWORD env variable.'),
    ]

    def __init__(self, *args, **kwargs):
        """
        Initializes the publish command, which includes building and validating
        newly built source and binary distribution files.
        """
        super(PublishCommand, self).__init__(*args, **kwargs)
        run_setuppy(['clean', 'man', 'sdist', 'bdist_wheel'])
        run_twine(['check', 'dist/*']) or fail(
            'Could not check packages in dist/*')
        self._username = None
        self._password = None
        self.username = None
        self.password = None

    def initialize_options(self):
        """Initializes the CLI arguments for this command."""
        self.username = None
        self.password = None

    def finalize_options(self):
        """Finalizes the username and password credentials."""

        self._username = self.username \
            if self.username \
            else getenv('PYPI_USERNAME', None)
        if not self._username:
            fail('Must specify username, either --username or $PYPI_USERNAME')

        self._password = self.password \
            if self.password \
            else getenv('PYPI_PASSWORD', None)
        if not self._password:
            fail('Must specify username, either --password or $PYPI_PASSWORD')

    def run(self):
        """
        Publishes the built source and binary distributions to the public PyPi
        repository.
        """
        run_twine(['upload',
                   '--verbose',
                   '--username', self._username,
                   '--password', self._password,
                   'dist/*']) \
            or fail('Cannot upload source/binary distribution.')


setup(name=PROJECT_NAME,
      version=PROJECT_VERSION,

      author='Mike Durso',
      author_email='rbprogrammer@gmail.com',
      classifiers=[
          'Development Status :: 5 - Production/Stable',
          'Environment :: Console',
          'Intended Audience :: Developers',
          'Intended Audience :: System Administrators',
          'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
          'Natural Language :: English',
          'Operating System :: MacOS',
          'Operating System :: POSIX',
          'Operating System :: POSIX :: Linux',
          'Operating System :: Unix',
          'Operating System :: Microsoft :: Windows',
          'Programming Language :: Python',
          'Programming Language :: Python :: 3',
          'Topic :: Utilities',
      ],
      cmdclass={
          'clean': CleanCommand,
          'man': BuildManPagesCommand,
          'publish': PublishCommand,
          'twine_check': TwineCheckCommand,
      },
      data_files=BuildManPagesCommand.get_data_files(),
      description='Another XDG Base Directory Specification utility.',
      entry_points={
          'console_scripts': [
              'xdg-env = xdgenvpy.__main__:main',
          ],
      },
      install_requires=load_requirements('requirements.txt'),
      long_description=read_repo_file('README.md'),
      long_description_content_type='text/markdown',
      packages=get_xdgenvpy_packages(),
      scripts=['bin/xdg-env-completion.bash'],
      tests_require=load_requirements('requirements-test.txt'),
      url='https://gitlab.com/deliberist/xdgenvpy',
      )



"""
Helper module that defines the default XDG Base Directory values.

This is designed to minimize system specific logic, effectively allowing for
Linux CI systems, like the Gitlab runners, to test as much code coverage as
possible.  This should help maximize the possibility that the Gitlab CI will
identify bugs rather than the AppVeyor environments.  The AppVeyor CI
environment is configured to test MacOS and Windows systems.
"""

from collections import defaultdict
from ntpath import pathsep as pathsep_nt
from os import getenv
from os import pathsep
from pathlib import Path
from platform import system
from posixpath import pathsep as pathsep_posix

# Determine the user's name.  If it cannot be determined through the environment
# variable, then default to something/anything.  This is standard practice for
# Unix systems, and Windows systems usually do the same.
_USER_NAME = getenv('USERNAME', 'xdgenvpy')

# Attempt to determine the user's ID number.  Not that the getuid() is a Unix
# only method.  So on Windows, default the ID number to just the user's name.
try:
    # pylint: disable=ungrouped-imports
    from os import getuid

    _USER_ID = getuid()
except ImportError:
    _USER_ID = _USER_NAME


def __unsupported_system():
    """
    :raises RuntimeError with a message identifying the unsupported system.
    """
    system_type = system()
    raise RuntimeError(f'Unknown platform: {system_type}')


# Mapping between system type and a path prefix when paths are relative.
_PREFIX_PATHS = defaultdict(__unsupported_system)
if not _PREFIX_PATHS:
    _PREFIX_PATHS.update({
        'Linux': Path('~').expanduser(),
        'Darwin': Path('~').expanduser(),
        'Windows': Path(getenv('APPDATA',
                               f'C:\\Users\\{_USER_NAME}\\AppData\\Roaming')),
    })

# Mapping between system type and multiple directory path separators.
_PATH_SEPARATORS = defaultdict(__unsupported_system)
if not _PATH_SEPARATORS:
    _PATH_SEPARATORS.update({
        'Linux': pathsep_posix,
        'Darwin': pathsep_posix,
        'Windows': pathsep_nt,
    })


def _get_system_default(linux, mac, windows, multiple_dirs, system_type):
    """
    Depending on the system type, this method will take the system specified
    directories, split them up based on the system's path separator, and finally
    return an immutable sequence of paths.

    :param str linux: The Linux default value.
    :param str mac: The Mac OS default value.
    :param str windows: The Windows default value.
    :param bool multiple_dirs: Flag indicating if normalization should process
        the default value as a path separated list of multiple directories.
    :param str system_type: The system type as reported by :meth:`os.system()`.

    :rtype: tuple
    :returns: A sequence of directories based on the system type.
    """
    dirs = defaultdict(__unsupported_system)
    dirs.update({
        'Linux': linux,
        'Darwin': mac,
        'Windows': windows,
    })
    dirs = dirs.get(system_type)

    if multiple_dirs:
        path_sep = _PATH_SEPARATORS[system_type]
        dirs = dirs.split(path_sep)

    if isinstance(dirs, str):
        dirs = tuple([dirs])

    return tuple(dirs)


def _normalize(linux, mac, windows, multiple_dirs):
    """
    Given the defaults for Linux, Mac OS, and Windows based systems, this
    function determines which system specific default value to use, normalize
    the paths with path prefixes, and lastly it will potentially combine
    multiple directories into a single string.

    :param str linux: The Linux default value.
    :param str mac: The Mac OS default value.
    :param str windows: The Windows default value.
    :param bool multiple_dirs: Flag indicating if normalization should process
        the default value as a path separated list of multiple directories.

    :rtype: str
    :returns: A string containing a normalized default value for the current
        system.
    """
    system_type = system()
    prefix_path = _PREFIX_PATHS[system_type]
    dirs = _get_system_default(linux, mac, windows, multiple_dirs, system_type)
    dirs = [prefix_path.joinpath(d).as_posix() for d in dirs]
    normalized = str(_PATH_SEPARATORS[system_type]).join(dirs)
    return normalized


def system_path_separator():
    """ :returns: The platform specific path separator. """
    return pathsep


def XDG_DATA_HOME():
    """
    :rtype: str
    :returns: The spec defined value for the :code:`XDG_DATA_HOME` variable.
    """
    return _normalize(linux='.local/share',
                      mac='.local/share',
                      windows='local/share',
                      multiple_dirs=False)


def XDG_CONFIG_HOME():
    """
    :rtype: str
    :returns: The spec defined value for the :code:`XDG_CONFIG_HOME` variable.
    """
    return _normalize(linux='.config',
                      mac='.config',
                      windows='config',
                      multiple_dirs=False)


def XDG_CACHE_HOME():
    """
    :rtype: str
    :returns: The spec defined value for the :code:`XDG_CACHE_HOME` variable.
    """
    return _normalize(linux='.cache',
                      mac='.cache',
                      windows='cache',
                      multiple_dirs=False)


def XDG_RUNTIME_DIR():
    """
    :rtype: str
    :returns: The spec defined value for the :code:`XDG_RUNTIME_DIR` variable.
    """
    return _normalize(linux=f'/run/user/{_USER_ID}',
                      mac=f'/tmp/run/user/{_USER_ID}',
                      windows=f'run/user/{_USER_NAME}',
                      multiple_dirs=False)


def XDG_DATA_DIRS():
    """
    :rtype: str
    :returns: The spec defined value for the :code:`XDG_DATA_DIRS` variable.
    """
    return _normalize(linux='/usr/local/share:/usr/share',
                      mac='/usr/local/share:/usr/share',
                      windows='usr/share/',
                      multiple_dirs=True)


def XDG_CONFIG_DIRS():
    """
    :rtype: str
    :returns: The spec defined value for the :code:`XDG_CONFIG_DIRS` variable.
    """
    return _normalize(linux='/etc/xdg',
                      mac='/etc/xdg',
                      windows='etc/xdg',
                      multiple_dirs=True)



"""
Various specifications specify files and file formats. This specification
defines where these files should be looked for by defining one or more base
directories relative to which files should be located.

The XDG Base Directory Specification is based on the following concepts:

- There is a single base directory relative to which user-specific data files
  should be written. This directory is defined by the environment variable
  :code:`$XDG_DATA_HOME`.

- There is a single base directory relative to which user-specific configuration
  files should be written. This directory is defined by the environment variable
  :code:`$XDG_CONFIG_HOME`.

- There is a set of preference ordered base directories relative to which data
  files should be searched. This set of directories is defined by the
  environment variable :code:`$XDG_DATA_DIRS`.

- There is a set of preference ordered base directories relative to which
  configuration files should be searched. This set of directories is defined by
  the environment variable :code:`$XDG_CONFIG_DIRS`.

- There is a single base directory relative to which user-specific non-essential
  (cached) data should be written. This directory is defined by the environment
  variable :code:`$XDG_CACHE_HOME`.

- There is a single base directory relative to which user-specific runtime files
  and other file objects should be placed. This directory is defined by the
  environment variable :code:`$XDG_RUNTIME_DIR`.

All paths set in these environment variables must be absolute. If an
implementation encounters a relative path in any of these variables it should
consider the path invalid and ignore it.
"""

from os import getenv
from os import makedirs
from pathlib import Path

from xdgenvpy import _defaults


class XDG():
    """
    A top level XDG object that basically just implements the XDG specification.
    """

    def __init__(self):
        """Initializes the base XDG object."""
        self._base_dir = Path('~').expanduser()

    @property
    def XDG_DATA_HOME(self):
        """
        There is a single base directory relative to which user-specific data
        files should be written. This directory is defined by the environment
        variable :code:`$XDG_DATA_HOME`.

        :code:`$XDG_DATA_HOME` defines the base directory relative to which user
        specific data files should be stored. If :code:`$XDG_DATA_HOME` is
        either not set or empty, a default equal to :code:`$HOME/.local/share`
        should be used.

        :rtype: str
        :return: A string containing the value of :code:`$XDG_DATA_HOME`.
        """
        return getenv('XDG_DATA_HOME', _defaults.XDG_DATA_HOME())

    @property
    def XDG_CONFIG_HOME(self):
        """
        There is a single base directory relative to which user-specific
        configuration files should be written. This directory is defined by the
        environment variable :code:`$XDG_CONFIG_HOME`.

        :code:`$XDG_CONFIG_HOME` defines the base directory relative to which
        user specific configuration files should be stored. If
        :code:`$XDG_CONFIG_HOME` is either not set or empty, a default equal to
        :code:`$HOME/.config` should be used.

        :rtype: str
        :return: A string containing the value of :code:`$XDG_CONFIG_HOME`.
        """
        return getenv('XDG_CONFIG_HOME', _defaults.XDG_CONFIG_HOME())

    @property
    def XDG_CACHE_HOME(self):
        """
        There is a single base directory relative to which user-specific
        non-essential (cached) data should be written. This directory is defined
        by the environment variable :code:`$XDG_CACHE_HOME`.

        code:`$XDG_CACHE_HOME` defines the base directory relative to which user
        specific non-essential data files should be stored. If
        :code:`$XDG_CACHE_HOME` is either not set or empty, a default equal to
        :code:`$HOME/.cache` should be used.

        :rtype: str
        :return: A string containing the value of :code:`$XDG_CACHE_HOME`.
        """
        return getenv('XDG_CACHE_HOME', _defaults.XDG_CACHE_HOME())

    @property
    def XDG_RUNTIME_DIR(self):
        """
        There is a single base directory relative to which user-specific runtime
        files and other file objects should be placed. This directory is defined
        by the environment variable :code:`$XDG_RUNTIME_DIR`.

        :code:`$XDG_RUNTIME_DIR` defines the base directory relative to which
        user-specific non-essential runtime files and other file objects (such
        as sockets, named pipes, ...) should be stored. The directory MUST be
        owned by the user, and he MUST be the only one having read and write
        access to it. Its Unix access mode MUST be 0700.

        The lifetime of the directory MUST be bound to the user being logged in.
        It MUST be created when the user first logs in and if the user fully
        logs out the directory MUST be removed. If the user logs in more than
        once he should get pointed to the same directory, and it is mandatory
        that the directory continues to exist from his first login to his last
        logout on the system, and not removed in between. Files in the directory
        MUST not survive reboot or a full logout/login cycle.

        The directory MUST be on a local file system and not shared with any
        other system. The directory MUST by fully-featured by the standards of
        the operating system. More specifically, on Unix-like operating systems
        AF_UNIX sockets, symbolic links, hard links, proper permissions, file
        locking, sparse files, memory mapping, file change notifications, a
        reliable hard link count must be supported, and no restrictions on the
        file name character set should be imposed. Files in this directory MAY
        be subjected to periodic clean-up. To ensure that your files are not
        removed, they should have their access time timestamp modified at least
        once every 6 hours of monotonic time or the 'sticky' bit should be set
        on the file.

        If :code:`$XDG_RUNTIME_DIR` is not set applications should fall back to
        a replacement directory with similar capabilities and print a warning
        message. Applications should use this directory for communication and
        synchronization purposes and should not place larger files in it, since
        it might reside in runtime memory and cannot necessarily be swapped out
        to disk.

        :rtype: str
        :return: A string containing the value of :code:`$XDG_RUNTIME_DIR`.
        """
        return getenv('XDG_RUNTIME_DIR', _defaults.XDG_RUNTIME_DIR())

    @property
    def XDG_DATA_DIRS(self):
        """
        There is a set of preference ordered base directories relative to which
        data files should be searched. This set of directories is defined by the
        environment variable :code:`$XDG_DATA_DIRS`.

        :code:`$XDG_DATA_DIRS` defines the preference-ordered set of base
        directories to search for data files in addition to the
        :code:`$XDG_DATA_HOME` base directory. The directories in
        :code:`$XDG_DATA_DIRS` should be seperated with a colon ':'.

        If :code:`$XDG_DATA_DIRS` is either not set or empty, a value equal to
        :code:`/usr/local/share/:/usr/share/` should be used.

        The order of base directories denotes their importance; the first
        directory listed is the most important. When the same information is
        defined in multiple places the information defined relative to the more
        important base directory takes precedent. The base directory defined by
        :code:`$XDG_DATA_HOME` is considered more important than any of the base
        directories defined by :code:`$XDG_DATA_DIRS`.

        :rtype: tuple
        :return: A sequence containing the value of :code:`$XDG_DATA_DIRS`.
        """
        data_home = getenv('XDG_DATA_HOME', _defaults.XDG_DATA_HOME())
        data_dirs = getenv('XDG_DATA_DIRS', _defaults.XDG_DATA_DIRS())
        return data_home + _defaults.system_path_separator() + data_dirs

    @property
    def XDG_CONFIG_DIRS(self):
        """
        There is a set of preference ordered base directories relative to which
        configuration files should be searched. This set of directories is
        defined by the environment variable code:`$XDG_CONFIG_DIRS`.

        :code:`$XDG_CONFIG_DIRS` defines the preference-ordered set of base
        directories to search for configuration files in addition to the
        :code:`$XDG_CONFIG_HOME` base directory. The directories in
        :code:`$XDG_CONFIG_DIRS` should be seperated with a colon ':'.

        If :code:`$XDG_CONFIG_DIRS` is either not set or empty, a value equal to
        :code:`/etc/xdg` should be used.

        The order of base directories denotes their importance; the first
        directory listed is the most important. When the same information is
        defined in multiple places the information defined relative to the more
        important base directory takes precedent. The base directory defined by
        :code:`$XDG_CONFIG_HOME` is considered more important than any of the
        base directories defined by :code:`$XDG_CONFIG_DIRS`.

        :rtype: tuple
        :return: A sequence containing the value of :code:`$XDG_CONFIG_DIRS`.
        """
        config_home = getenv('XDG_CONFIG_HOME', _defaults.XDG_CONFIG_HOME())
        config_dirs = getenv('XDG_CONFIG_DIRS', _defaults.XDG_CONFIG_DIRS())
        return config_home + _defaults.system_path_separator() + config_dirs


class XDGPackage(XDG):
    """
    An instance of this class helps build package specific directories according
    to the XDG specification.  In the case of:

        -   :code:`$XDG_DATA_DIR`
        -   :code:`$XDG_CONFIG_DIR`
        -   :code:`$XDG_CACHE_DIR`

    variables, this is accomplished by simply appending the package name to the
    directory locations.  All other variables essentially remain unchanged,
    unless they require the DATA/CONFIG/CACHE package directories.
    """

    def __init__(self, package_name):
        """
        Initializes the object with the specified package name.

        :param str package_name: The name of the package.
        """
        super(XDGPackage, self).__init__()
        if package_name:
            package_name = package_name.strip()
        if not package_name:
            raise ValueError('Package name must be specified.')
        self._package_name = package_name

    @property
    def XDG_DATA_HOME(self):
        """:return: A package directory relative to the XDG data directory."""
        base = super(XDGPackage, self).XDG_DATA_HOME
        return Path().joinpath(base, self._package_name)

    @property
    def XDG_CONFIG_HOME(self):
        """:return: A package directory relative to the XDG config directory."""
        base = super(XDGPackage, self).XDG_CONFIG_HOME
        return Path().joinpath(base, self._package_name)

    @property
    def XDG_CACHE_HOME(self):
        """:return: A package directory relative to the XDG cache directory."""
        base = super(XDGPackage, self).XDG_CACHE_HOME
        return Path().joinpath(base, self._package_name)

    @property
    def XDG_RUNTIME_DIR(self):
        """
        :return: A package directory relative to the XDG runtime directory.
        """
        base = super(XDGPackage, self).XDG_RUNTIME_DIR
        return Path().joinpath(base, self._package_name)


class XDGPedanticPackage(XDGPackage):
    """
    An instance of this class goes one step further than the :class:`XDGPackage`
    in that it ensures package-specific directories exist in the file system.
    If one of the

        -   :code:`$XDG_DATA_DIR`
        -   :code:`$XDG_CONFIG_DIR`
        -   :code:`$XDG_CACHE_DIR`

    does not exist, the getter method will ensure the directory (and all parent
    directories) exist.  This same logic is not applied to non-package specific
    directories as they can often contain system level directories.  And thus
    may not be writable by the current user.
    """

    @staticmethod
    def _safe_path(path):
        """
        Ensures the supplied path exists.  If the path does not exist then the
        directory and all parent directories will be created.

        :param str path: The path to ensure exists.

        :rtype: str
        :return: The supplied string, but ensures the path exists.
        """
        if not Path(path).exists():
            makedirs(path)
        return path

    @property
    def XDG_DATA_HOME(self):
        """
        :return: A package directory relative to the XDG data directory with the
                guarantee that the directory exists.
        """
        base = super(XDGPedanticPackage, self).XDG_DATA_HOME
        return XDGPedanticPackage._safe_path(base)

    @property
    def XDG_CONFIG_HOME(self):
        """
        :return: A package directory relative to the XDG config directory with
                the guarantee that the directory exists.
        """
        base = super(XDGPedanticPackage, self).XDG_CONFIG_HOME
        return XDGPedanticPackage._safe_path(base)

    @property
    def XDG_CACHE_HOME(self):
        """
        :return: A package directory relative to the XDG cache directory with
                the guarantee that the directory exists.
        """
        base = super(XDGPedanticPackage, self).XDG_CACHE_HOME
        return XDGPedanticPackage._safe_path(base)

    @property
    def XDG_RUNTIME_DIR(self):
        """
        :return: A package directory relative to the XDG runtime directory with
                the guarantee that the directory exists.
        """
        base = super(XDGPedanticPackage, self).XDG_RUNTIME_DIR
        return XDGPedanticPackage._safe_path(base)
